from logging import getLogger
from datetime import datetime

from ..dbio import engine
from ..dbio import database
from ..dbio import table
from ..dbio import view
from ..dbio import trigger

from ..models.project import Project
from ..models.mapping import Mapping
# from ..models import const

logger = getLogger(__name__)


# def _load_for_entry(proj: Project, map: Mapping, ins: str, tbl: str, dm: DataModel):
#     # 自身のデータをロードし、存在しなければ親を再帰的に遡ってロードする。
#     items = load_all_data(proj.folder, map.id, ins, tbl)
#     if items is None:
#         if map.parent is not None:
#             return _load_for_entry(proj, map.parent, ins, tbl, dm)
#     # extendの処理
#         return None
#     for item in items:
#         for col, val in item.items():
#             if col in dm.settings:
#                 typ = dm.settings[col]['type']
#                 if typ in [const.BIND_TYPE_BLANK, const.BIND_TYPE_REFS]:
#                     continue
#                 bind = proj.bindings[typ]
#                 if bind is None:
#                     continue
#                 if bind.type == const.BIND_TYPE_EXTEND:
#                     module = importlib.import_module(bind.value)
#                     work = val if val is not None else ''
#                     result = module.convert(proj, map, ins, tbl, dm, *work.split(','))
#                     item[col] = result
#     return items


# def _load_data_model(proj: Project, map: Mapping, ins: str, tbl: str) -> bool | None:
#     if is_exist_data_model(proj.folder, map.id, ins, tbl) is False:
#         if map.parent is None:
#             return None
#         return _load_data_model(proj, map.parent, ins, tbl)

#     dm = load_model(
#         get_data_model_name(proj.folder, map.id, ins, tbl),
#         DataModel,
#         id=map.id,
#         instance=ins,
#         table_name=tbl
#     )
#     return dm


class Operation:

    @staticmethod
    def get_instance(folder: str, env: str, database: str, deploy: str):
        project = Project.load(folder)
        return Operation(project, env, database, deploy)

    def __init__(self, project: Project, env: str, database: str, deploy: str):
        self.project = project
        self.environ = project.envs[env]

        # FIXME マッピングの取得？データベースから特定されるもの
        # self.map = mapping.get(project.folder, env)

        self.conn = engine.get_connection(self.environ.deployment[deploy])
        self.ymd = datetime.strftime(datetime.now(), '%Y%m%d%H%M%S')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.conn.close()

    def create_database(self, map: Mapping, all: str):
        if all == 'drop':
            logger.info(f'database {map.name}')
            if database.is_exist(self.conn, map.name):
                database.drop(self.conn, map.name)
            database.create(self.conn, map.name)
        else:
            # 差分更新または個別指定の場合で、データベースが存在しない場合は作成する。
            if not database.is_exist(self.conn, map.name):
                logger.info(f'database {map.name} was created.')
                database.create(self.conn, map.name)

    def create_table(self, map: Mapping, all: str, target: str):
        schema = map.build_schema(self.project.schemas, self.environ.schemas)

        for tbl in schema.tables:
            if not all and target != tbl.table_name:
                continue
            # テーブルが存在しない場合は作成する。
            if not table.is_exist(self.conn, map.name, tbl):
                logger.info(f'table {map.name}.{tbl.table_name} was created.')
                table.create(self.conn, map.name, tbl)
                continue
            # データのバックアップ
            logger.info(f'backup {map.name}.{tbl.table_name}')
            table.backup(self.conn, map.name, tbl, self.ymd)
            # テーブルの再作成
            logger.info(f'drop & create table {map.name}.{tbl.table_name}')
            table.drop(self.conn, map.name, tbl)
            table.create(self.conn, map.name, tbl)

        for vw in schema.views:
            if not all and target != vw.view_name:
                continue
            # ビューが存在しない場合は作成する。
            if not view.is_exist_view(self.conn, map.name, vw):
                logger.info(f'view {map.name}.{vw.view_name} was created.')
                view.create_view(self.conn, map.name, vw)
            else:
                # ビューの再作成
                logger.info(f'drop & create view {map.name}.{vw.view_name}')
                view.drop_view(self.conn, map.name, vw)
                view.create_view(self.conn, map.name, vw)

        for tr in schema.triggers:
            if not all and target != tr.trigger_name:
                continue
            # トリガーが存在しない場合は作成する。
            if not trigger.is_exist_trigger(self.conn, map.name, tr):
                logger.info(f'trigger {map.name}.{tr.trigger_name} was created.')
                trigger.create_trigger(self.conn, map.name, tr)
            else:
                # トリガーの再作成
                logger.info(f'drop & create trigger {map.name}.{tr.trigger_name}')
                trigger.drop_trigger(self.conn, map.name, tr)
                trigger.create_trigger(self.conn, map.name, tr)

    def insert_data(self, all: bool, target: str):
        # データ投入
        for ins in self.map.instances:
            schema = self.project.schemas[ins]
            for tbl in schema.tables.values():
                if not all and target != tbl.table_name:
                    continue
                # dm = _load_data_model(self.project, self.map, ins, tbl.table_name)
                # items = _load_for_entry(self.project, self.map, ins, tbl.table_name, dm)
                # if items is not None:
                #     logger.info(f'insert {self.map.id}.{tbl.table_name}')
                #     table.insert(self.conn, self.map.id, tbl, items)
                # if dm is not None and dm.sync_mode == 'update_diff':
                #     # 新規作成時にはバックアップはない
                #     if table.is_exist_backup(self.conn, self.map.id, tbl, self.ymd):
                #         # バックアップからデータを復元
                #         # create_tableの中で再作成されているため
                #         logger.info(f'restore {self.map.id}.{tbl.table_name}')
                #         table.restore(self.conn, self.map.id, tbl, self.ymd)
                # engine.commit(self.conn)

    def reset_all(self):
        for map in self.environ.databases:
            self.create_database(map, 'drop')
            self.create_table(map, 'drop', None)

    def require(self, instance: str, table_name: str):
        # ユニットテストなどで指定したデータを挿入する。
        # schema = self.project.schemas[instance]
        # tbl = schema.get_table(table_name)
        # dm = _load_data_model(self.project, self.map, instance, table_name)
        # items = _load_for_entry(self.project, self.map, instance, table_name, dm)
        # if items is not None:
        #     logger.info(f'insert {self.map.id}.{table_name}')
        #     table.insert(self.conn, self.map.id, tbl, items)
        pass


def apply(project, env: str, database: str, target: str, all: str, deploy: str):
    with Operation(project, env, database, deploy) as op:
        for map in op.environ.databases:
            if database is not None and map.name != database:
                continue
            # データベースの作成
            op.create_database(map, all)
            op.create_table(map, all, target)
        # op.insert_data(all, target)



import pydantic

from .base import BaseSchema
from .column import Column
from .column import ColumnManager
from .index import Index
from .index import IndexManager
from .relation import Relation
from .relation import RelationManager
from .notes import Note
from .notes import NoteManager


class MySQLTableOptions(BaseSchema):
    """MySQL-specific table options"""
    # ストレージエンジン
    engine: str = "InnoDB"  # InnoDB, MyISAM, MEMORY, etc.

    # 文字セット・照合順序
    charset: str | None = None  # utf8mb4, latin1, etc.
    collation: str | None = None  # utf8mb4_unicode_ci, etc.

    # AUTO_INCREMENT
    auto_increment: int | None = None  # 開始値

    # 行フォーマット
    row_format: str | None = None  # DYNAMIC, COMPRESSED, REDUNDANT, COMPACT

    # パーティション（MySQL）
    partition_by: str | None = None  # RANGE, LIST, HASH, KEY
    partition_expression: str | None = None
    partition_count: int | None = None  # HASH/KEY用のパーティション数


class Table(BaseSchema):
    instance: str = pydantic.Field(exclude=True)
    table_name: str = pydantic.Field(exclude=True)
    display_name: str
    columns_: list[Column] = pydantic.Field(default_factory=list, alias='columns')
    indexes_: list[Index] = pydantic.Field(default_factory=list, alias='indexes')
    relations_: list[Relation] = pydantic.Field(default_factory=list, alias='relations')
    notes_: list[Note] = pydantic.Field(default_factory=list, alias='notes')

    # MySQL固有のテーブルオプション
    mysql_options: MySQLTableOptions | None = None

    @property
    def columns(self) -> list[Column]:
        return ColumnManager(self.columns_)

    @property
    def indexes(self) -> list[Index]:
        return IndexManager(self.indexes_)

    @property
    def relations(self) -> list[Relation]:
        return RelationManager(self.relations_)

    @property
    def notes(self) -> NoteManager:
        return NoteManager(self.notes_)


class TableManager:

    def __init__(self, tables: dict[str, Table]):
        self.tables = tables

    def __getitem__(self, table_name: str) -> Table:
        return self.tables[table_name]

    def __iter__(self):
        yield from self.tables.values()

    def __len__(self) -> int:
        return len(self.tables)

    def __contains__(self, table_name: str) -> bool:
        return table_name in self.tables

    def add(self, table: Table) -> None:
        if table.table_name in self.tables:
            raise ValueError(f"Table '{table.table_name}' already exists")
        self.tables[table.table_name] = table

    def remove(self, table_name: str) -> None:
        if table_name not in self.tables:
            raise KeyError(f"Table '{table_name}' not found")
        del self.tables[table_name]



import pydantic
import yaml
import os

from .base import BaseSchema
from .mapping import Mapping
from .exceptions import DBGearEntityExistsError
from .exceptions import DBGearEntityNotFoundError
from ..utils.populate import auto_populate_from_keys


class DatabaseInfo(BaseSchema):
    name: str
    database: str
    description: str | None = None
    active: bool = True


class TenantConfig(BaseSchema):
    name: str = pydantic.Field(exclude=True)
    ref: str

    # tenant variables
    databases: list[DatabaseInfo] = []


class TenantRegistry(BaseSchema):
    """Registry of tenant configurations"""
    folder: str = pydantic.Field(exclude=True)
    name: str = pydantic.Field(exclude=True)
    tenants: dict[str, TenantConfig] = pydantic.Field(default_factory=dict)

    @classmethod
    def load(cls, folder: str, name: str):
        path = f'{folder}/{name}/tenant.yaml'
        if not os.path.exists(path):
            return None
        with open(path, 'r', encoding='utf-8') as f:
            data = yaml.safe_load(f)
        populated_data = auto_populate_from_keys(data, {
            'tenants.$1.name': '$1',
        })
        return cls(
            folder=folder,
            name=name,
            **populated_data
        )

    def save(self) -> None:
        """Save tenant configurations to a YAML file"""
        with open(f'{self.folder}/{self.name}/tenant.yaml', 'w', encoding='utf-8') as f:
            yaml.dump(
                self.model_dump(
                    by_alias=True,
                    exclude_none=True,
                    exclude_defaults=True
                ),
                f,
                indent=2,
                allow_unicode=True,
                default_flow_style=False,
                sort_keys=False)

    def __getitem__(self, name: str) -> TenantConfig:
        return self.tenants[name]

    def __iter__(self):
        yield from self.tenants.values()

    def __len__(self) -> int:
        return len(self.tenants)

    def __contains__(self, name: str) -> bool:
        return name in self.tenants

    def add(self, tenant: TenantConfig) -> None:
        if tenant.name in self.tenants:
            raise DBGearEntityExistsError(f'Tenant {tenant.name} already exists')
        self.tenants[tenant.name] = tenant

    def remove(self, name: str) -> None:
        if name not in self.tenants:
            raise DBGearEntityNotFoundError(f'Tenant {name} does not exist')
        del self.tenants[name]

    def materialize(self):
        for tenant in self.tenants.values():
            map = Mapping.load(self.folder, self.name, tenant.ref)

            for database in tenant.databases:
                if not database.active:
                    continue
                clone = map.model_copy(deep=True)
                clone.name = database.database
                yield clone



"""
DBGear unified exception classes
"""


class DBGearError(Exception):
    """Base exception class for all DBGear model errors"""
    pass


class DBGearEntityExistsError(DBGearError):
    """Raised when attempting to create an entity that already exists"""
    pass


class DBGearEntityNotFoundError(DBGearError):
    """Raised when attempting to access an entity that does not exist"""
    pass


class DBGearEntityRemovalError(DBGearError):
    """Raised when an entity cannot be removed due to constraints"""
    pass



"""
Generic schema importer with dynamic module loading.

This module provides a generic interface for importing schemas from various formats
by dynamically loading specific importer modules.
"""

import importlib
from pathlib import Path
from typing import Any


def import_schema(importer_type: str, *args, **kwargs) -> Any:
    """
    Import schema using the specified importer module.

    Dynamically loads the importer module and calls its retrieve function.

    Args:
        importer_type: Name of the importer module (e.g., 'a5sql_mk2')
        *args, **kwargs: Arguments to pass to the retrieve function

    Returns:
        SchemaManager: The imported schema manager object

    Raises:
        ImportError: If the importer module cannot be found
        AttributeError: If the importer module doesn't have a retrieve function
    """
    try:
        # Dynamically import the importer module
        module = importlib.import_module(f'dbgear.importers.{importer_type}')

        # Get and execute the retrieve function
        if hasattr(module, 'retrieve'):
            return module.retrieve(*args, **kwargs)
        else:
            raise AttributeError(f"Module '{importer_type}' does not have 'retrieve' function")

    except ImportError as e:
        raise ImportError(f"Importer '{importer_type}' not found: {e}")


def list_importers() -> list[str]:
    """
    Get a list of available importer modules.

    Returns:
        List of available importer module names
    """
    importers_dir = Path(__file__).parent.parent / 'importers'
    if not importers_dir.exists():
        return []

    return [f.stem for f in importers_dir.glob('*.py') if f.stem != '__init__']



from . import engine
from .templates.mysql import template_engine


def is_exist(conn, database):
    sql = template_engine.render('mysql_check_database_exists')
    result = engine.select_one(conn, sql, {'database_name': database})
    return result is not None


def create(conn, database, charset='utf8mb4', collation='utf8mb4_unicode_ci'):
    sql = template_engine.render(
        'mysql_create_database',
        database_name=database,
        charset=charset,
        collation=collation)
    engine.execute(conn, sql)


def drop(conn, database):
    sql = template_engine.render('mysql_drop_database', database_name=database)
    engine.execute(conn, sql)


# def grant(conn, database, user, host, privileges='ALL PRIVILEGES'):
#     sql = f'GRANT {privileges} ON {database}.* TO :user'
#     print(sql)
#     engine.execute(conn, sql, {'user': f'{user}@{host}'})


# def create_user(conn, user, host, password):
#     sql = f"CREATE USER IF NOT EXISTS '{user}'@'{host}' IDENTIFIED BY '{password}'"
#     engine.execute(conn, sql)



import pydantic

from .base import BaseSchema
from .notes import Note
from .notes import NoteManager


class Trigger(BaseSchema):
    """Database trigger definition"""
    instance: str = pydantic.Field(exclude=True)
    trigger_name: str = pydantic.Field(exclude=True)
    display_name: str
    table_name: str  # 対象テーブル名
    timing: str  # BEFORE | AFTER | INSTEAD OF
    event: str  # INSERT | UPDATE | DELETE
    condition: str | None = None  # WHEN条件（オプション）
    body: str  # トリガーの実行内容
    notes_: list[Note] = pydantic.Field(default_factory=list, alias='notes')

    @property
    def notes(self) -> NoteManager:
        return NoteManager(self.notes_)


class TriggerManager:

    def __init__(self, triggers: dict[str, Trigger]):
        self.triggers = triggers

    def __getitem__(self, trigger_name: str) -> Trigger:
        return self.triggers[trigger_name]

    def __iter__(self):
        yield from self.triggers.values()

    def __len__(self) -> int:
        return len(self.triggers)

    def __contains__(self, trigger_name: str) -> bool:
        return trigger_name in self.triggers

    def add(self, trigger: Trigger) -> None:
        if trigger.trigger_name in self.triggers:
            raise ValueError(f"Trigger '{trigger.trigger_name}' already exists")
        self.triggers[trigger.trigger_name] = trigger

    def remove(self, trigger_name: str) -> None:
        if trigger_name not in self.triggers:
            raise KeyError(f"Trigger '{trigger_name}' does not exist")
        del self.triggers[trigger_name]



import pydantic
import yaml
import pathlib
import os

from .base import BaseSchema
from .schema import Schema
from .schema import SchemaManager
from .exceptions import DBGearEntityExistsError
from .exceptions import DBGearEntityNotFoundError
from .exceptions import DBGearEntityRemovalError


class SharedInfo(BaseSchema):
    environ: str
    mapping: str


class Mapping(BaseSchema):
    folder: str = pydantic.Field(exclude=True)
    environ: str = pydantic.Field(exclude=True)
    name: str = pydantic.Field(exclude=True)
    description: str
    schemas: list[str] = []
    shared: SharedInfo | None = None
    deploy: bool = False

    @classmethod
    def load(cls, folder: str, environ: str, name: str):
        with open(f'{folder}/{environ}/{name}/_mapping.yaml', 'r', encoding='utf-8') as f:
            data = yaml.safe_load(f)
        return cls(
            folder=folder,
            environ=environ,
            name=name,
            **data
        )

    def build_schema(self, project_schema: SchemaManager, environ_schema: SchemaManager | None):
        schema = Schema(name=self.name)
        for name in self.schemas:
            if name in project_schema:
                schema.update(project_schema[name])
            if environ_schema is not None and name in environ_schema:
                schema.update(environ_schema[name])
        return schema


class MappingManager:

    def __init__(self, folder: str, environ: str):
        self.folder = folder
        self.environ = environ

    def __getitem__(self, key: str) -> Mapping:
        return Mapping.load(self.folder, self.environ, key)

    def __iter__(self):
        for path in sorted(pathlib.Path(self.folder).glob(f'{self.environ}/*/_mapping.yaml')):
            name = str(path.parent.relative_to(pathlib.Path(self.folder) / self.environ))
            yield Mapping.load(self.folder, self.environ, name)

    def add(self, mapping: Mapping) -> None:
        path = os.path.join(self.folder, mapping.environ, mapping.name)
        if os.path.exists(path):
            raise DBGearEntityExistsError(f'Mapping {mapping.name} already exists in {self.folder}/{mapping.environ}')

        os.makedirs(path, exist_ok=True)
        with open(os.path.join(path, '_mapping.yaml'), 'w', encoding='utf-8') as f:
            yaml.dump(
                mapping.model_dump(
                    by_alias=True,
                    exclude_none=True,
                    exclude_defaults=True
                ),
                f,
                indent=2,
                allow_unicode=True,
                default_flow_style=False,
                sort_keys=False)

    def remove(self, name: str) -> None:
        path = os.path.join(self.folder, self.environ, name)
        if not os.path.exists(path):
            raise DBGearEntityNotFoundError(f'Mapping {name} does not exist in {self.folder}/{self.environ}')
        files = [f for f in os.listdir(path) if f != '_mapping.yaml']
        if files:
            raise DBGearEntityRemovalError(f'Cannot remove {path}: files other than _mapping.yaml exist')
        os.remove(os.path.join(path, '_mapping.yaml'))
        os.rmdir(path)



import pydantic

from .base import BaseSchema
from .notes import Note
from .notes import NoteManager


class ViewColumn(BaseSchema):
    """View column definition (auto-generated from SQL parsing)"""
    column_name: str
    display_name: str
    column_type: str
    nullable: bool
    source_table: str | None = None  # 参照元テーブル
    source_column: str | None = None  # 参照元カラム
    is_computed: bool = False  # 計算列かどうか


class View(BaseSchema):
    """Database view definition"""
    instance: str = pydantic.Field(exclude=True)
    view_name: str = pydantic.Field(exclude=True)
    display_name: str
    select_statement: str
    notes_: list[Note] = pydantic.Field(default_factory=list, alias='notes')

    # 以下は将来のSQL解析で自動生成される予定
    _parsed_columns: list[ViewColumn] = []  # SQL解析結果をキャッシュ
    _dependencies: list[str] = []  # 参照テーブル/ビューを自動検出
    _is_parsed: bool = False  # 解析済みフラグ

    @property
    def notes(self) -> NoteManager:
        return NoteManager(self.notes_)


class ViewManager:

    def __init__(self, views: dict[str, View]):
        self.views = views

    def __getitem__(self, view_name: str) -> View:
        return self.views[view_name]

    def __iter__(self):
        yield from self.views.values()

    def __len__(self) -> int:
        return len(self.views)

    def __contains__(self, view_name: str) -> bool:
        return view_name in self.views

    def add(self, view: View) -> None:
        if view.view_name in self.views:
            raise ValueError(f"View '{view.view_name}' already exists")
        self.views[view.view_name] = view

    def remove(self, view_name: str) -> None:
        if view_name not in self.views:
            raise KeyError(f"View '{view_name}' does not exist")
        del self.views[view_name]



# from importlib import import_module
import pydantic
import yaml

from .base import BaseSchema
from .schema import SchemaManager
from .environ import EnvironManager
from .option import DBGearOptions


# class Binding(BaseSchema):
#     type: str
#     value: str


class Project(BaseSchema):
    folder: str = pydantic.Field(exclude=True)
    project_name: str
    description: str
    options: DBGearOptions = pydantic.Field(default_factory=DBGearOptions)

    _schemas: SchemaManager | None = None

    # def __init__(self, folder: str, data: dict):
    #     self.folder = folder
    #     self.schemas = SchemaManager.load(f'{folder}/schema.yaml')
    #     self.project_name = data['project_name']
    #     self.description = data['description']
    # self._definitions = data.get('definitions', [])
    # self._rules = data['rules']
    # self._bindings = {k: Binding(**v) for k, v in data['bindings'].items()}
    # self.deployments = data['deployments']

    @classmethod
    def load(cls, folder: str):
        with open(f'{folder}/project.yaml', 'r', encoding='utf-8') as f:
            data = yaml.safe_load(f)
        mgr = cls(
            folder=folder,
            **data
        )
        return mgr

    def save(self) -> None:
        with open(f'{self.folder}/project.yaml', 'w', encoding='utf-8') as f:
            yaml.dump(
                self.model_dump(
                    by_alias=True,
                    exclude_none=True,
                    exclude_defaults=True
                ),
                f,
                indent=2,
                allow_unicode=True,
                default_flow_style=False,
                sort_keys=False
            )

    @property
    def schemas(self) -> SchemaManager:
        if self._schemas is None:
            self._schemas = SchemaManager.load(f'{self.folder}/schema.yaml')
        return self._schemas

    @property
    def envs(self) -> EnvironManager:
        return EnvironManager(self.folder)

    # @property
    # def bindings(self) -> dict[str, Binding]:
    #     return self._bindings

    # @property
    # def rules(self) -> dict[str, str]:
    #     return self._rules

    # @property
    # def schemas(self) -> SchemaManager:
    #     return self._schemas

    # @property
    # def instances(self) -> list[str]:
    #     return list(self.schemas.get_schemas().keys())

    # def read_definitions(self) -> None:
    #     for items in self._definitions:
    #         self.logger.info(f"definition: {items['type']}")
    #         module = import_module(f'.{items["type"]}', 'dbgear.core.definitions')
    #         result = module.retrieve(self.folder, **items)

    #         self._schemas.update(result)


# def project(request) -> Project:
#     return request.app.state.project



import pydantic
import yaml
import pathlib
import os

from typing import Any

from .base import BaseSchema
from .exceptions import DBGearEntityExistsError
from .exceptions import DBGearEntityRemovalError


class SettingInfo(BaseSchema):
    type: str
    width: int | None = None
    environ: str | None = None
    schema_name: str | None = None
    table_name: str | None = None


class DataSource:
    folder: str
    environ: str
    schema_name: str
    table_name: str
    segment: str | None = None
    data: list[dict[str, Any]]

    @property
    def filename(self) -> str:
        if self.segment:
            return f"{self.schema_name}@{self.table_name}#{self.segment}.dat"
        return f"{self.schema_name}@{self.table_name}.dat"

    def load(self) -> None:
        path = os.path.join(self.folder, self.environ, self.filename)
        if not os.path.exists(path):
            raise FileNotFoundError(f"Data source file {path} does not exist.")
        with open(path, 'r', encoding='utf-8') as f:
            self.data = yaml.safe_load(f)

    def save(self) -> None:
        path = os.path.join(self.folder, self.environ, self.filename)
        with open(path, 'w', encoding='utf-8') as f:
            yaml.dump(
                self.data,
                f,
                indent=2,
                allow_unicode=True,
                default_flow_style=False,
                sort_keys=False
            )


class DataModel(BaseSchema):
    folder: str = pydantic.Field(exclude=True)
    environ: str = pydantic.Field(exclude=True)
    schema_name: str
    table_name: str
    description: str
    layout: str
    settings: dict[str, SettingInfo]
    sync_mode: str
    value: str | None = None
    caption: str | None = None
    segment: str | None = None
    x_axis: str | None = None
    y_axis: str | None = None
    cells: list[str] | None = None

    @classmethod
    def load(cls, folder: str, environ: str, name: str):
        with open(os.path.join(folder, environ, name), 'r', encoding='utf-8') as f:
            data = yaml.safe_load(f)
        return cls(
            folder=folder,
            environ=environ,
            **data
        )

    def save(self) -> None:
        filename = os.path.join(self.folder, self.environ, self.filename)
        with open(filename, 'w', encoding='utf-8') as f:
            yaml.dump(
                self.model_dump(
                    by_alias=True,
                    exclude_none=True,
                    exclude_defaults=True
                ),
                f,
                indent=2,
                allow_unicode=True,
                default_flow_style=False,
                sort_keys=False
            )

    @property
    def filename(self) -> str:
        return f"{self.schema_name}@{self.table_name}.yaml"

    @property
    def datasources(self):
        if not self.segment:
            yield DataSource(
                folder=self.folder,
                environ=self.environ,
                schema_name=self.schema_name,
                table_name=self.table_name,
            )

        for path in sorted(pathlib.Path(self.folder, self.environ, self.schema_name).glob(f"{self.schema_name}@{self.table_name}#*.dat")):
            seg = path.stem.split('#', 1)[1] if '#' in path.stem else None
            yield DataSource(
                folder=self.folder,
                environ=self.environ,
                schema_name=self.schema_name,
                table_name=self.table_name,
                segment=seg
            )


class DataModelManager:

    def __init__(self, folder: str, environ: str):
        self.folder = folder
        self.environ = environ

    def __getitem__(self, key: str) -> DataModel:
        return DataModel.load(self.folder, self.environ, key)

    def __iter__(self):
        for path in sorted(pathlib.Path(self.folder).glob(f'{self.environ}/*/*.yaml')):
            if path.name == '_mapping.yaml':
                continue
            name = str(path.parent.relative_to(pathlib.Path(self.folder) / self.environ))
            yield DataModel.load(self.folder, self.environ, name)

    def add(self, model: DataModel) -> None:
        filename = os.path.join(self.folder, self.environ, model.filename)
        if os.path.exists(filename):
            raise DBGearEntityExistsError(f'DataModel {model.filename} already exists in {self.folder}/{self.environ}')
        with open(filename, 'w', encoding='utf-8') as f:
            yaml.dump(
                model.model_dump(
                    by_alias=True,
                    exclude_none=True,
                    exclude_defaults=True
                ),
                f,
                indent=2,
                allow_unicode=True,
                default_flow_style=False,
                sort_keys=False)

    def remove(self, model: DataModel) -> None:
        filename = os.path.join(self.folder, self.environ, model.filename)
        try:
            os.remove(filename)
        except OSError as e:
            raise DBGearEntityRemovalError(f'Failed to remove DataModel {model.filename}: {e}')



import pydantic

from .base import BaseSchema


class DBGearOptions(BaseSchema):
    """Common options for Project and Environment configurations"""

    # Database construction options
    create_foreign_key_constraints: bool = pydantic.Field(
        default=True,
        description="Whether to create foreign key constraints during database construction"
    )

    # Future options can be added here
    # create_indexes: bool = True
    # validate_data_integrity: bool = True
    # enable_triggers: bool = True
